'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var sagaEffects = require('redux-saga/effects');
var invariant = _interopDefault(require('invariant'));
var redux = require('redux');
var createSagaMiddleware = _interopDefault(require('redux-saga'));

const slash = '/';

/**
 * Add a prefix like `namespace/setName`
 *
 * @param {*} namespace
 */
function addSetPrefix (namespace) {
  return name => (namespace ? `${namespace}${slash}set${upperCaseFirst(name)}`
    : `set${upperCaseFirst(name)}`)
}

/**
 * Add a prefix like `namespace/name`
 *
 * @param {*} namespace
 */
function addPrefix (namespace) {
  return namespace ? name => `${namespace}${slash}${name}` : name => name
}

function upperCaseFirst (s) {
  if (s.length > 1) {
    return s[0].toUpperCase() + s.substring(1, s.length)
  } else {
    return s.toUpperCase()
  }
}

const thunkMiddleware = ({dispatch, getState}) => next => action => {
  if (typeof action === 'function') {
    return action(dispatch, getState)
  }
  return next(action)
};

const defaultConfig = {
  models: {},
  middlewares: [],
  enableThunk: true,
  devtools: {
    enable: true,
    options: {}
  }
};
const mergeObject = (origin, replacer) => {
  return replacer ? { ...origin, ...(replacer || {}) } : origin || {}
};

function mergeConfig (config) {
  if (!config) return defaultConfig
  return {
    models: mergeObject(defaultConfig.models, config.models),
    middlewares: config.middlewares || [],
    // devtools: config.devtools ? config.devtools : defaultConfig.devtools
    devtools: mergeObject(defaultConfig.devtools, config.devtools),
    enableThunk: config.enableThunk !== false
  }
}

var helpers = {
  takeLatest: (fn, ...args) => key => {
    return function * () {
      yield sagaEffects.fork(function * () {
        yield sagaEffects.takeLatest(key, fn, ...args);
      });
    }
  },
  takeEvery: (fn, ...args) => key => {
    return function * () {
      yield sagaEffects.fork(function * () {
        yield sagaEffects.takeEvery(key, fn, ...args);
      });
    }
  },
  throttle: (fn, ...args) => key => {
    const ms = args[0];
    const restArgs = args.slice(1);
    return function * () {
      yield sagaEffects.fork(function * () {
        yield sagaEffects.throttle(ms, key, fn, restArgs);
      });
    }
  }
};

// Argument is an Object but not array
// About typeof null === 'object' See http://2ality.com/2013/10/typeof-null.html
const isNotArrayObject = a => typeof a === 'object' && !Array.isArray(a);
const isNotNullObject = a => isNotArrayObject(a) && a !== null;

// Return whether the object has the given key
const includeKey = (obj, key) => (isNotArrayObject(obj) ? Object.keys(obj).includes(key) : false);

// The 'obj' and 'payloads' should be objects but not arrays
// and only compare properties in the first level.
const includeNewKeys = (obj, ...payloads) => {
  const originKeys = Object.keys(obj);
  const allKeys = Object.keys(obj).concat(payloads.reduce((keys, p) => keys.concat(Object.keys(p)), []));
  return originKeys.length !== new Set(allKeys).size
};

/* eslint-disable no-underscore-dangle */

class Sirius {
  constructor (config) {
    this.config = mergeConfig(config);
    this._models = [];
    this._effects = [];
    this._sagas = [];
  }

  /**
   * Add model dynamically
   *
   * @param {*} model
   */
  addModel (model) {
    invariant(typeof model.namespace === 'string', 'model `namespace` field is required');
    checkModel(model);
    this._store.replaceReducer(
      mergeReducers(this._reducers, {
        [model.namespace]: createRootReducer(model, model.namespace)
      })
    );
    getSagas.apply(this, [model, model.namespace]).forEach(this.runSaga);
    this._models.push(model);
  }
  /**
   * Create the redux store
   *
   */
  store () {
    // each sirius instance should only have one store
    invariant(!this._store, 'Only support one store');
    const config = this.config;
    const reducers = {};
    for (const name of Object.keys(config.models)) {
      const model = config.models[name];
      checkModel(model);
      reducers[name] = createRootReducer(model, name);
      getSagas.apply(this, [model, name]);
      this._models.push({
        namespace: name,
        ...model
      });
    }
    let store;
    const sagaMiddleware = createSagaMiddleware();
    const { middlewares } = this.config;
    let mws;
    if (!Array.isArray(middlewares)) {
      mws = redux.applyMiddleware(sagaMiddleware);
    } else {
      if (config.enableThunk) {
        middlewares.push(thunkMiddleware);
      }
      mws = redux.applyMiddleware(...middlewares, sagaMiddleware);
    }
    this._reducers = reducers;
    const combinedReducer = mergeReducers(reducers);
    // eslint-disable-next-line no-undef
    if (__DEV__ && this.config.devtools.enable) {
      store = redux.createStore(combinedReducer,
        // redux devtools support
        window.__REDUX_DEVTOOLS_EXTENSION__ && window.__REDUX_DEVTOOLS_EXTENSION__(this.config.devtools.options),
        mws);
    } else {
      store = redux.createStore(combinedReducer, mws);
    }
    this._sagas.forEach(sagaMiddleware.run);
    this._store = store;
    this.runSaga = sagaMiddleware.run;
    return store
  }
}

function mergeReducers (reducers, newReducers) {
  const finalyReduers = { ...reducers, ...newReducers };
  if (!Object.keys(finalyReduers).length) {
    return state => state
  }
  return redux.combineReducers(finalyReduers)
}

function checkModel (model) {
  invariant(Object.keys(model).includes('state'), 'model `state` field is required');
  return model
}

function getSagas (model, name) {
  const sagas = [];
  const _effects = model.effects;
  if (!_effects) {
    return []
  }
  // Normally we define sagas with String patterns in helpers like `takeLatest`.
  // But there are senarioes when we need native sagas with complex patterns.
  // See https://redux-saga.js.org/docs/api/#takeeverypattern-saga-args about partterns
  if (typeof _effects === 'function') {
    const effects = _effects(helpers);
    for (const key of Object.keys(effects) || []) {
      const sagaKey = addPrefix(name)(key);
      sagas.push(effects[key](sagaKey));
    }
  } else if (Array.isArray(_effects)) {
    for (const s of _effects || []) {
      sagas.push(function * e () {
        yield sagaEffects.fork(s);
      });
    }
  } else if (typeof _effects === 'object') {
    // Object effects must follow the pattern `{ default, native }` which
    // 'default' represents for the functional model-scoped sirius sagas
    // and 'native' represents for the native redux sagas
    if (_effects.default && typeof _effects.default) {
      const effects = _effects.default(helpers);
      for (const key of Object.keys(effects) || []) {
        const sagaKey = addPrefix(name)(key);
        sagas.push(effects[key](sagaKey));
      }
    }
    if (_effects.native && Array.isArray(_effects.native)) {
      for (const s of _effects.native || []) {
        sagas.push(function * e () {
          yield sagaEffects.fork(s);
        });
      }
    }
  } else {
    invariant(false, 'Effects must be array, function or object');
  }
  this._sagas = this._sagas.concat(sagas);
  return sagas
}

/**
 * Create a root reducer based on model state and model reducers
 *
 * If you have a model like the below
 * {
 *   namespace: 'form',
 *   state: {
 *     loading: false,
 *     password: ''
 *   }
 * }
 *
 * then sirius will generate a reducer for each state field following a preset rule:
 *
 * for state loading : form/setLoading
 * for state password : form/setPassword
 *
 * And a reducer receving action whose type is '<namespace>/merge' (in the senario above is 'form/merge')
 * will be generated.
 * If the state is an Object (not Array, not null), the action payload must be a 'sub-object' of the state
 * which means all the properties of payload can be also found in the state. the 'merge' reducer will
 * do a 'merge-like' action to the state just like Object Spreading.
 *
 * If the state is not an Object, reducer replace the state with payload directly.
 *
 * @param {*} model
 * @param {*} name
 */
function createRootReducer (model, name) {
  const handlers = {};
  // auto-generate reducers
  if (isNotNullObject(model.state)) {
    for (const key of Object.keys(model.state)) {
      handlers[addSetPrefix(name)(key)] = (state, action) => (includeKey(action, 'payload') ? { ...state, [key]: action.payload } : state);
    }
  }
  // reducer for updating multiple field of the state in one action
  // or replace the state directly
  handlers[addPrefix(name)('merge')] = (state, action) => {
    if (!includeKey(action, 'payload')) {
      return state
    }
    const payload = action.payload;
    if (isNotArrayObject(state)) {
      if (includeNewKeys(state, payload)) {
        return state
      } else {
        return { ...state, ...payload }
      }
    } else {
      return payload
    }
  };
  // user defined reducers
  for (const r of Object.keys(model.reducers || {})) {
    const reducer = model.reducers[r];
    let finalReducer;
    if (typeof reducer === 'function') {
      finalReducer = (state, action) => reducer(state, action);
    } else {
      finalReducer = (state) => state;
    }
    // notice the reducer override ocurrs here
    handlers[addPrefix(name)(r)] = finalReducer;
  }
  return (state = model.state, action) => (handlers[action.type] ? handlers[action.type](state, action) : state)
}

const effects = sagaEffects;

exports.effects = effects;
exports.default = Sirius;
